#!/usr/bin/env python
#
# Copyright (c) 2015 Serguei Glavatski ( verser  from cnc-club.ru )
# Copyright (c) 2020 Probe Screen NG Developers
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

import math
import os
import sys
import time
from datetime import datetime
from functools import wraps
from subprocess import PIPE, Popen

import gtk
import linuxcnc
import hal
import hal_glib
import pango

import pylab as pl
import numpy as np

from .configparserl import ProbeScreenConfigParser
from .util import restore_task_mode
from six import _meth_self
from pip._internal.utils.outdated import SELFCHECK_DATE_FMT

CONFIGPATH1 = os.environ["CONFIG_DIR"]

class ProbeScreenBase(object):
    # --------------------------
    #
    #  INIT
    #
    # --------------------------
    def __init__(self, halcomp, builder, useropts):
        self.builder = builder
        self.halcomp = halcomp

        # Load the machines INI file
        self.inifile = linuxcnc.ini(os.environ["INI_FILE_NAME"])
        if not self.inifile:
            self.error_dialog("Error, no INI File given")

        # Load Probe Screen Preferences
        self.prefs = ProbeScreenConfigParser(self.get_preference_file_path())

        # Which display is in use? AXIS / gmoccapy / unknown
        self.display = self.get_display() or "unknown"

        # LinuxCNC Command / Stat / Error Interfaces
        self.command = linuxcnc.command()
        self.stat = linuxcnc.stat()
        self.stat.poll()
        self.error = linuxcnc.error_channel()
        self.error.poll()

        # History Area
        textarea = builder.get_object("textview1")
        self.buffer = textarea.get_property("buffer")

        # Warning Dialog
        self.window = builder.get_object("window1")
        

        # VCP Reload Action
        self._vcp_action_reload = self.builder.get_object("vcp_action_reload")
        
        self._vcp_action_mdi = self.builder.get_object("vcp_action_mdi")

        # Results Display
        self._lb_probe_xp = self.builder.get_object("lb_probe_xp")
        self._lb_probe_yp = self.builder.get_object("lb_probe_yp")
        self._lb_probe_xm = self.builder.get_object("lb_probe_xm")
        self._lb_probe_ym = self.builder.get_object("lb_probe_ym")
        self._lb_probe_lx = self.builder.get_object("lb_probe_lx")
        self._lb_probe_ly = self.builder.get_object("lb_probe_ly")
        self._lb_probe_z = self.builder.get_object("lb_probe_z")
        self._lb_probe_d = self.builder.get_object("lb_probe_d")
        self._lb_probe_xc = self.builder.get_object("lb_probe_xc")
        self._lb_probe_yc = self.builder.get_object("lb_probe_yc")
        self._lb_probe_a = self.builder.get_object("lb_probe_a")
        
        #tool editor
        temp = self.inifile.find("EMCIO", "TOOL_TABLE")
        tooledit = builder.get_object("tooledit1")
        tooledit.set_filename(temp)
        
        # pins used to read and write linuxcnc program memory
        try:
            self.halcomp.newpin("hal_read_memory", hal.HAL_FLOAT, hal.HAL_IN)
            self.halcomp.newpin("hal_write_memory", hal.HAL_FLOAT, hal.HAL_OUT)
        except ValueError:
            print "foo"
            # ignore duplicate pin errors


    # --------------------------
    #
    #  MDI Command Methods
    #
    # --------------------------
    #@restore_task_mode
    def gcode(self, s, data=None):
        self.switch_to_mdi()
        
        gcodes = self.stat.gcodes
        
        for l in s.split("\n"):
            # Search for G1 followed by a space, otherwise we'll catch G10 too.
            if "G1 " in l:
                l += " F#<_ini[TOOLSENSOR]RAPID_SPEED>"
            self.command.mdi(l)
            #self.add_history_text(l)
            # wait for the line to finish
            self.wait_complete_non_blocking(50, "gcode execution")
            self.error_poll(gcodes)

        return 0

    #@restore_task_mode
    def ocode(self, s, data=None):
        self.switch_to_mdi()
        
        # capture gcode state for restore
        gcodes = self.stat.gcodes
        
        self.command.mdi(s)
        
        self.stat.poll()
        while self.stat.interp_state != linuxcnc.INTERP_IDLE:
            self.stat.poll()
            self.error_poll(gcodes)
            self.command.wait_complete(0.02)
            while gtk.events_pending():
                gtk.main_iteration_do(True)
        self.wait_complete_non_blocking(50, "ocode execution")
        # restore g codes just in case there was an error
        #self.restore_gcodes(gcodes)
        self.ocode_return_read()
        
    
    def switch_to_mdi(self):
        self.stat.poll()
        if self.stat.task_mode != linuxcnc.MODE_MDI:
            self.command.mode(linuxcnc.MODE_MDI)
            self.wait_complete_non_blocking(5, "switching to MDI mode")

    def wait_complete_non_blocking(self, timeout = 5, msg = ""):
        i = 0
        while i < (timeout * 50):
            if self.command.wait_complete(0.020) == -1:
                while gtk.events_pending():
                    gtk.main_iteration_do(True)
                i += 1
            else:
                break
        if i > (timeout * 50):
            raise ValueError("Timeout waiting for interpreter in: " + msg)

    def restore_gcodes(self, gcodes):
        s = ""
        for code in gcodes:
            # don't restore G0 or G1 as they should be re-set, also could be more than one 0 in the array
            if code != -1 and code != 0 and code != 10 and code != 30:
                s += "G%.1f " % float(float(code) / 10.0)
        #self.gcode(s)
        
    
    def error_poll(self, gcodes):
        
        error_pin = self.halcomp["gui_error"]
        err = self.error.poll()
        
        if err:
            kind, text = err
            if kind in (linuxcnc.NML_ERROR, linuxcnc.OPERATOR_ERROR):
                #self.command.abort()
                #self.command.reset_interpreter()
                #self.restore_gcodes(gcodes)
                # send the error back to the linuxcnc queue
                #  TODO: parse kind to return the correct error type
                self.command.mdi("(ABORT, " + text + ")\n")
                raise ValueError("Error running command, emcerr:" + text)
            else:
                self.comman.mdi("msg, " + text + ")\n")
                #self.add_history_text("Message during gcode: " + text)
                
        # just in case the GUI poll caught the error
        if error_pin:
            #self.command.abort()
            #self.command.reset_interpreter()
            #self.restore_gcodes(gcodes)
            text = "See notification popup"
            raise ValueError("UI indicated an error in gcode: " + text)


    def ocode_return_read(self):
        retval = self.halcomp["hal_read_memory"]
        
        if retval != 0:
            # error in o code fuction
            if retval == -1:
                s = "Search distance must be greater than zero"
            elif retval == -2:
                s = "Z angle must be between 0 and 90"
            elif retval == -3:
                s = "Search velocity must be greater than zero"
            elif retval == -4:
                s = "Probe velocity must be greater than zero"
            elif retval == -5:
                s = "Probe hardware error"
            elif retval == -6:
                s = "Probe low battery"
            elif retval == -7:
                s = "Probe tripped before G38.3 move"
            elif retval == -8:
                s = "Probe failed to make contact during probe move"
            elif retval == -10:
                s = "ocode did not complete unexpectedly"
            else:
                s = "unknown error code " + retval
            
            raise ValueError(s)
                

    def ngc274_memory_read(self, address):
        
        if type(address) is int:
            # is a fixed memory location read
            if address < 31 or address > 5601:
                raise ValueError("ngc274 memory read address must be between 31 and 5601")
            s = "#" + str(address)
        elif type(address) is str:
            if str[0] != "_":
                raise ValueError("ngc274 memory read must target a global variable")
            # is a relative memory location read
            s = "#<" + address + ">"
        else:
            raise ValueError("ngc274 memory read address must be of type str or int")
        
        pin = str(self.inifile.find("PROBE_SCREEN", "NGC274_READ_PIN"))
        
        self.gcode("M68 E" + pin + " Q[" + s + "]")
        
        value = self.halcomp["hal_read_memory"]
        
        return value
        
    def ngc274_memory_write(self, address, value):
        if type(address) is int:
            # is a fixed memory location write
            if address < 31 or address > 5601:
                raise ValueError("ngc274 memory write address must be between 31 and 5601")
            s = "#" + str(address)
        elif type(address) is str:
            if str[0] != "_":
                raise ValueError("ngc274 memory write must target a global variable")
            # is a relative memory location write
            s = "#<" + address + ">"
        else:
            raise ValueError("ngc274 memory write address must be of type str or int")
        
        self.halcomp["hal_write_memory"] = value
        
        pin = str(self.inifile.find("PROBE_SCREEN", "NGC274_WRITE_PIN"))
        
        self.gcode("M66 E" + pin + " L0\n " + s + " = #5399\n")



    # --------------------------
    #
    #  Utility Methods
    #
    # --------------------------
    def get_display(self):
        # gmoccapy or axis ?
        temp = self.inifile.find("DISPLAY", "DISPLAY")
        if not temp:
            print(
                "****  PROBE SCREEN GET INI INFO **** \n Error recognition of display type : %s"
                % temp
            )
        return temp

    def get_preference_file_path(self):
        # we get the preference file, if there is none given in the INI
        # we use toolchange2.pref in the config dir
        temp = self.inifile.find("DISPLAY", "PREFERENCE_FILE_PATH")
        if not temp:
            machinename = self.inifile.find("EMC", "MACHINE")
            if not machinename:
                temp = os.path.join(CONFIGPATH1, "probe_screen.pref")
            else:
                machinename = machinename.replace(" ", "_")
                temp = os.path.join(CONFIGPATH1, "%s.pref" % machinename)
        print("****  probe_screen GETINIINFO **** \n Preference file path: %s" % temp)
        return temp

    def vcp_reload(self):
        """ Realods the VCP - e.g. after changing changing changing origin/zero points """
        self._vcp_action_reload.emit("activate")

    # --------------------------
    #
    #  History and Logging Methods
    #
    # --------------------------
    def add_history(
        self,
        tool_tip_text,
        s="",
        xm=None,
        xc=None,
        xp=None,
        lx=None,
        ym=None,
        yc=None,
        yp=None,
        ly=None,
        z=None,
        d=None,
        a=None,
    ):
        c = "{0: <10}".format(tool_tip_text)
        if "Xm" in s:
            c += "X-=%.4f " % xm
            self._lb_probe_xm.set_text("%.4f" % xm)
        if "Xc" in s:
            c += "Xc=%.4f " % xc
            self._lb_probe_xc.set_text("%.4f" % xc)
        if "Xp" in s:
            c += "X+=%.4f " % xp
            self._lb_probe_xp.set_text("%.4f" % xp)
        if "Lx" in s:
            c += "Lx=%.4f " % lx
            self._lb_probe_lx.set_text("%.4f" % lx)
        if "Ym" in s:
            c += "Y-=%.4f " % ym
            self._lb_probe_ym.set_text("%.4f" % ym)
        if "Yc" in s:
            c += "Yc=%.4f " % yc
            self._lb_probe_yc.set_text("%.4f" % yc)
        if "Yp" in s:
            c += "Y+=%.4f " % yp
            self._lb_probe_yp.set_text("%.4f" % yp)
        if "Ly" in s:
            c += "Ly=%.4f " % ly
            self._lb_probe_ly.set_text("%.4f" % ly)
        if "Z" in s:
            c += "Z=%.4f " % z
            self._lb_probe_z.set_text("%.4f" % z)
        if "D" in s:
            c += "D=%.4f" % d
            self._lb_probe_d.set_text("%.4f" % d)
        if "A" in s:
            c += "Angle=%.3f" % a
            self._lb_probe_a.set_text("%.3f" % a)

        self.add_history_text(c)

    def add_history_text(self, text):
        # Prepend a timestamp to all History lines
        text = datetime.now().strftime("%H:%M:%S  ") + text

        # Remove the oldest history entries when we have a large
        # number of entries.
        i = self.buffer.get_end_iter()
        if i.get_line() > 1000:
            i.backward_line()
            self.buffer.delete(i, self.buffer.get_end_iter())

        # Add the line of text to the top of the history
        i.set_line(0)
        self.buffer.insert(i, "%s \n" % text)

    def _dialog(
        self, gtk_type, gtk_buttons, message, secondary=None, title=_("Probe Screen NG")
    ):
        """ displays a dialog """
        dialog = gtk.MessageDialog(
            self.window,
            gtk.DIALOG_DESTROY_WITH_PARENT,
            gtk_type,
            gtk_buttons,
            message,
        )
        # if there is a secondary message then the first message text is bold
        if secondary:
            dialog.format_secondary_text(secondary)
        dialog.set_keep_above(True)
        dialog.show_all()
        dialog.set_title(title)
        responce = dialog.run()
        dialog.destroy()
        return responce == gtk.RESPONSE_OK

    def warning_dialog(self, message, secondary=None, title=_("Probe Screen NG")):
        """ displays a warning dialog """
        return self._dialog(
            gtk.MESSAGE_WARNING, gtk.BUTTONS_OK, message, secondary, title
        )

    def error_dialog(self, message, secondary=None, title=_("Probe Screen NG")):
        """ displays a warning dialog and exits the probe screen"""
        self._dialog(gtk.MESSAGE_ERROR, gtk.BUTTONS_CLOSE, message, secondary, title)
        sys.exit(1)

    def display_result_a(self, value):
        # TODO: Convert the remaining uses of this to add_history and remove.
        self._lb_probe_a.set_text("%.3f" % value)

    
    
    # --------------------------
    #
    #  Generic Probe Movement Methods
    #
    # --------------------------
    def z_clearance_down(self, z_offset = 0):
        # move Z - z_clearance
        return """G91
        G1 Z-%f
        G90
        """ % (
            self.halcomp["ps_z_clearance"] + z_offset
        )

    def z_clearance_up(self, z_offset = 0):
        # move Z + z_clearance
        return """G91
        G1 Z%f
        G90
        """ % (
            self.halcomp["ps_z_clearance"] + z_offset
        )
    
    def move_z(self, z):
        return """G90
        G1 Z%f
        """ % z
    
    def move_x(self, x):
        return """G90
        G1 X%f
        """ % x
    
    def move_y(self, y):
        return """G90
        G1 Y%f
        """ % y
    
    def move_xy(self, x, y):
        return """G90
        G1 X%f Y%f
        """ % (x, y)
    
    def move_xyz(self, x, y, z):
        return """G90
        G1 X%f Y%f Z%f
        """ % (x, y, z)
        
    def move_rel_xy(self, x, y):
        return """G91
        G1 X%f Y%f
        G90
        """ % (x, y)
    
    def spindle_orient(self, angle):
        # TODO: Add if statement to disable via ini
        return """M19 R%.4f
        G4 P1
        """ % angle
        
    def probe_on(self, data=None):
        if self.halcomp["probe_is_on"] != 1:
            self.gcode("""M64 P0
                G4 P%f
                """ % float(self.inifile.find("PROBE_SCREEN", "PROBE_ON_TIME"))
            )
        else:
            self.gcode("""M64 P0
                """
            )
        if self.halcomp["probe_low_batt"] == 1:
            self.add_history_text("Probe battery is low")
            raise ValueError("Probe battery is low")
        if self.halcomp["probe_error"] == 1:
            self.add_history_text("Probe hardware error")
            raise ValueError("Probe hardware error")
        
    def probe_off(self, data=None):
        self.gcode("""M65 P0
            """
        )
    
    
    def probe_raw(self, xy_angle = 0, z_angle = 0):
        self.ocode("o<psng2_probe> call [%f] [%f] [%f] [%f] [%f] [%f] [%f]" %
            ( self.halcomp["ps_probe_max"],
            self.halcomp["ps_probe_latch"],
            xy_angle,
            z_angle,
            self.halcomp["ps_searchvel"],
            self.halcomp["ps_probevel"],
            float(self.inifile.find("PROBE_SCREEN", "PROBE_SETTLE"))
            ))
        a = self.probed_position_with_offsets()
        
        # return xyz position
        return float(a[0]), float(a[1]), float(a[2])
        
    def probe_fine(self, xy_angle = 0, z_angle = 0):
        self.ocode("o<psng2_probe_fine> call [%f] [%f] [%f] [%f] [%f] [%f]" % (
            self.halcomp["ps_probe_latch"],
            xy_angle,
            z_angle,
            self.halcomp["ps_searchvel"],
            self.halcomp["ps_probevel"],
            float(self.inifile.find("PROBE_SCREEN", "PROBE_SETTLE"))
            ))
        a = self.probed_position_with_offsets()
        
        return float(a[0]), float(a[1]), float(a[2])
        
        
    # direction is 1 for xp and -1 for xn
    # move_enable turns on the clearance moves, useful to disable in some circumstances
    # if zdown is false then just probe in the x axis and don't move up or down
    def probe_x(self, direction = 1, move_enable = True, zdown = True, orient_enable = True, raw_output = False, s = ""):
        
        if orient_enable:
            s += self.spindle_orient(0)
        
        # move X - xy_clearance
        if move_enable:
            s += self.move_rel_xy(direction * -1 * self.halcomp["ps_xy_clearance"], 0)
        if zdown:
            s += self.z_clearance_down()
        
        if s:
            self.gcode(s)
            s = ""
        
        if direction == 1:
            #self.ocode("o<psng_xplus> call")
            px, py, pz = self.probe_raw(0, 0)
        else:
            #self.ocode("o<psng_xminus> call")
            px, py, pz = self.probe_raw(180, 0)
        
        #a = self.probed_position_with_offsets()
        
        if zdown:
            s += self.z_clearance_up()
        
        #TODO: add probe calibration offset
        #raw_value = float(a[1])
        raw_value = px
        if raw_output:
            value = raw_value
        else:
            value = raw_value + 0.5 * direction * self.halcomp["ps_probe_diam"]
            
        if move_enable:
            s += self.move_x(value)
        
        if s:
            self.gcode(s)
        
        return value

    def probe_y(self, direction = 1, move_enable = True, zdown = True, orient_enable = True, raw_output = False, s = ""):
        
        if orient_enable:
            s += self.spindle_orient(0)
            
        # move Y - xy_clearance
        if move_enable:
            s += self.move_rel_xy(0, direction * -1 * self.halcomp["ps_xy_clearance"])
        if zdown:
            s += self.z_clearance_down()
        
        if s:
            self.gcode(s)
            s = ""
        
        if direction == 1:
            #self.ocode("o<psng_yplus> call")
            px, py, pz = self.probe_raw(90, 0)
        else:
            #self.ocode("o<psng_yminus> call")
            px, py, pz = self.probe_raw(270, 0)
        
        #a = self.probed_position_with_offsets()
        
        if zdown:
            s += self.z_clearance_up()
        
        #TODO: add probe calibration offset
        #raw_value = float(a[1])
        raw_value = py
        if raw_output:
            value = raw_value
        else:
            value = raw_value + 0.5 * direction * self.halcomp["ps_probe_diam"]
        
        if move_enable:
            s += self.move_y(value)
        
        if s:
            self.gcode(s)
        
        return value
    
    def probe_z(self, probe_depth):
        a = self.current_position_with_offsets()
        px, py, pz = self.probe_raw(0, 90)
        self.gcode(self.move_z(a[2]))
        #self.ocode("o<psng_probe_table> call [%.4f]" % probe_depth)
        #a = self.probed_position_with_offsets()
        return pz
    
    # probe must start in the rough center of the area to be measured
    # direction = 1 means this is an outside shape
    # direction = -1 means this is an inside shape
    # x,y_estimate are used to start the probing in the right place
    # if zdown is false, the probe is not moved up and down between touchoffs, ignored if direction = 1
    # orient_angle is used if orient_enable is true
    # raw_output, if true the calibration values are not applied to the results
    def probe_xy(self, 
        x_estimate, 
        y_estimate, 
        direction = 1, 
        z_offset = 0, 
        zdown = True, 
        move_center = False,
        orient_angle = 0, 
        orient_enable = True, 
        raw_output = False, 
        s = ""
    ):
        
        #get and store the starting point so we can move back there as needed
        b = self.current_position_with_offsets()
        
        if orient_enable:
            s += self.spindle_orient(orient_angle)
            
        # calculate clearances
        if direction == 1:
            xclear = x_estimate / 2 + self.halcomp["ps_xy_clearance"]
            yclear = y_estimate / 2 + self.halcomp["ps_xy_clearance"]
        else:
            if (x_estimate / 2) < self.halcomp["ps_xy_clearance"]:
                xclear = 0
            else:
                xclear = x_estimate / 2 - self.halcomp["ps_xy_clearance"]
            
            if (y_estimate / 2) < self.halcomp["ps_xy_clearance"]:
                yclear = 0
            else:
                yclear = y_estimate / 2 - self.halcomp["ps_xy_clearance"]
                
        s += self.move_rel_xy(xclear, 0)
        
        if direction == 1 or zdown:
            s += self.z_clearance_down(z_offset)
        
        self.gcode(s)
        s = ""
        
        #probe the x plus location
        if direction == 1:
            xp, yc, zp = self.probe_raw(180, 0)
            #self.ocode("o<psng_xminus> call")
        else:
            xp, yc, zp = self.probe_raw(0, 0)
            #self.ocode("o<psng_xplus> call")
        #a = self.probed_position_with_offsets()
        #xp = float(a[0])
        
        if direction == 1 or zdown:
            s += self.z_clearance_up(z_offset)
        
        #move to the next probing location
        s += self.move_x(b[0] - xclear)
        #s += self.move_rel_xy(-xclear, 0)
        
        if direction == 1 or zdown:
            s += self.z_clearance_down(z_offset)
        
        self.gcode(s)
        s = ""
        
        #probe the x minus location
        if direction == 1:
            xn, yc, zn = self.probe_raw(0, 0)
            #self.ocode("o<psng_xplus> call")
        else:
            xn, yc, zn = self.probe_raw(180, 0)
            #self.ocode("o<psng_xminus> call")
        #a = self.probed_position_with_offsets()
        #xn = float(a[0])
        
        if direction == 1 or zdown:
            s += self.z_clearance_up(z_offset)
        
        # return to center
        if move_center:
            s += self.move_xy((xp + xn) / 2, b[1] + yclear)
        else:
            s += self.move_xy(b[0], b[1] + yclear)
        
        if direction == 1 or zdown:
            s += self.z_clearance_down(z_offset)
        
        self.gcode(s)
        s = ""
        
        #probe the y plus location
        if direction == 1:
            xc, yp, zp = self.probe_raw(270, 0)
            #self.ocode("o<psng_yminus> call")
        else:
            xc, yp, zp = self.probe_raw(90, 0)
            #self.ocode("o<psng_yplus> call")
        #a = self.probed_position_with_offsets()
        #yp = float(a[1])
        
        if direction == 1 or zdown:
            s += self.z_clearance_up(z_offset)
        
        # return to center
        s += self.move_y(b[1] - yclear)
        
        if direction == 1 or zdown:
            s += self.z_clearance_down(z_offset)
        
        self.gcode(s)
        s = ""
        
        #probe the y minus location
        if direction == 1:
            xc, yn, zn = self.probe_raw(90, 0)
            #self.ocode("o<psng_yplus> call")
        else:
            xc, yn, zn = self.probe_raw(270, 0)
            #self.ocode("o<psng_yminus> call")
        #a = self.probed_position_with_offsets()
        #yn = float(a[1])
        
        if direction == 1 or zdown:
            s += self.z_clearance_up(z_offset)
        
        # move back to where we started
        if move_center:
            s += self.move_y((yp + yn) / 2)
        else:
            s += self.move_xy(b[0], b[1])
        
        self.gcode(s)
        
        #TODO: add probe calibration offset
        if not raw_output:
            xp += 0.5 * -1 * direction * self.halcomp["ps_probe_diam"]
            xn += 0.5 *  1 * direction * self.halcomp["ps_probe_diam"]
            yp += 0.5 * -1 * direction * self.halcomp["ps_probe_diam"]
            yn += 0.5 *  1 * direction * self.halcomp["ps_probe_diam"]
            
        return xp, xn, yp, yn
    
    # probe must start in the rough center of the area to be measured
    # direction = 1 means this is an outside shape
    # direction = -1 means this is an inside shape
    # x,y_estimate are used to start the probing in the right place
    # if zdown is false, the probe is not moved up and down between touchoffs, ignored if direction = 1
    # orient_angle is used if orient_enable is true
    # raw_output, if true the calibration values are not applied to the results
    def probe_circle(self, 
        dia_estimate, 
        num_points, 
        direction = 1, 
        z_offset = 0,
        zdown = True, 
        orient_angle = 0,
        num_samples = 1,
        orient_enable = True,
        fine_probe = False,
        raw_output = False, 
        s = ""
    ):
        if num_points < 2:
            raise ValueError("Must have at least 2 points to probe a circle")
        if num_samples < 1:
            raise ValueError("Must take at least one sample")
            
        #capture the center point for use
        centerpoint = self.current_position_with_offsets()
        
        if orient_enable:
            s += self.spindle_orient(orient_angle)
        
        # calculate clearances
        if fine_probe:
            backoff = self.halcomp["ps_probe_latch"] + self.halcomp["ps_probe_diam"] / 2
        else:
            backoff = self.halcomp["ps_xy_clearance"]
            
        if direction == 1:
            diaclear = dia_estimate / 2 + backoff
        else:
            if (dia_estimate / 2) < backoff:
                diaclear = 0
            else:
                diaclear = dia_estimate / 2 - backoff
                
        s += self.move_rel_xy(diaclear, 0)
        
        if direction == 1 or zdown:
            s += self.z_clearance_down(z_offset)
        
        self.gcode(s)
        s = ""
        
        radius = diaclear
        
        points = []
        
        for x in range(int(num_points)):
            a = self.current_position_with_offsets()
            current_angle = 360.0 / float(num_points) * float(x)
            probe_angle = current_angle
            if direction == 1:
                probe_angle += 180.0
                if probe_angle > 360.0:
                    probe_angle -= 360.0
            
            sval = "%.1f " % current_angle
            accumval = 0.0
            for i in range(int(num_samples)):
                if fine_probe:
                    xt, yt, zt = self.probe_fine(probe_angle, 0)
                else:
                    xt, yt, zt = self.probe_raw(probe_angle, 0)
                    
                #point = (xt, yt, zt)
                point = (xt, yt)
                length = ((xt - float(centerpoint[0])) ** 2 + (yt - float(centerpoint[1])) ** 2) ** .5
                if not raw_output:
                    length += 0.5 * -1 * direction * self.halcomp["ps_probe_diam"]
                sval += "(%.5f %.5f)" % (xt, yt)
                accumval += length
            length = accumval / float(num_samples)
            #lengths.append((probe_angle / 180.0 * math.pi, length))
            points.append(point)
            #self.add_history_text("angle: %.3f  length: %.5f" % (probe_angle, length))
            self.add_history_text(sval)
            # return to probe start point
            s = self.move_xyz(float(a[0]), float(a[1]), float(a[2]))
            
            # move in an arc to the next point
            if radius > 0:
                next_angle = 360.0 / float(num_points) * float(x + 1)
                next_x = math.cos(next_angle/180.0 * math.pi) * radius + float(centerpoint[0])
                next_y = math.sin(next_angle/180.0 * math.pi) * radius + float(centerpoint[1])
                i_val = float(centerpoint[0]) - float(a[0])
                j_val = float(centerpoint[1]) - float(a[1])
                s += "G3 X%f Y%f I%f J%f\n" % (next_x, next_y, i_val, j_val)
                
            self.gcode(s)
            s = ""
            
        if direction == 1 or zdown:
            s = self.z_clearance_up(z_offset)
        
        self.gcode(s)
        
        return points
        
    

    # --------------------------
    #
    #  Generic Position Calculations
    #
    # --------------------------
    def set_zerro(self, s="XYZ", x=0.0, y=0.0, z=0.0):
        if self.halcomp["set_zero"]:
            #  Z current position
            self.stat.poll()
            tmpz = (
                self.stat.position[2]
                - self.stat.g5x_offset[2]
                - self.stat.g92_offset[2]
                - self.stat.tool_offset[2]
            )
            c = "G10 L20 P0"
            s = s.upper()
            if "X" in s:
                x += self.halcomp["ps_offs_x"]
                c += " X%s" % x
            if "Y" in s:
                y += self.halcomp["ps_offs_y"]
                c += " Y%s" % y
            if "Z" in s:
                tmpz = tmpz - z + self.halcomp["ps_offs_z"]
                c += " Z%s" % tmpz
            self.gcode(c)
            self.vcp_reload()
            time.sleep(1)

            
    def current_position_with_offsets(self):
        self.stat.poll()
        positionval = list(self.stat.position)
        coord = list(self.stat.position)
        g5x_offset = list(self.stat.g5x_offset)
        g92_offset = list(self.stat.g92_offset)
        tool_offset = list(self.stat.tool_offset)

        for i in range(0, len(positionval) - 1):
            coord[i] = (
                positionval[i] - g5x_offset[i] - g92_offset[i] - tool_offset[i]
            )
        angl = self.stat.rotation_xy
        res = self._rott00_point(coord[0], coord[1], -angl)
        coord[0] = res[0]
        coord[1] = res[1]
        return coord
        
    def probed_position_with_offsets(self):
        self.stat.poll()
        probed_position = list(self.stat.probed_position)
        coord = list(self.stat.probed_position)
        g5x_offset = list(self.stat.g5x_offset)
        g92_offset = list(self.stat.g92_offset)
        tool_offset = list(self.stat.tool_offset)

        for i in range(0, len(probed_position) - 1):
            coord[i] = (
                probed_position[i] - g5x_offset[i] - g92_offset[i] - tool_offset[i]
            )
        angl = self.stat.rotation_xy
        res = self._rott00_point(coord[0], coord[1], -angl)
        coord[0] = res[0]
        coord[1] = res[1]
        return coord

    def _rott00_point(self, x1=0.0, y1=0.0, a1=0.0):
        """ rotate around 0,0 point coordinates """
        coord = [x1, y1]
        if a1 != 0:
            t = math.radians(a1)
            coord[0] = x1 * math.cos(t) - y1 * math.sin(t)
            coord[1] = x1 * math.sin(t) + y1 * math.cos(t)
        return coord

    def length_x(self, xm=None, xp=None):
        """ Calculates a length in the X direction """
        # Use previous value for xm if not supplied
        if xm is None:
            xm = self._lb_probe_xm.get_text()
            # Use None if no previous value exists
            if xm == "":
                xm = None
            else:
                xm = float(xm)

        # Use previous value for xp if not supplied
        if xp is None:
            xp = self._lb_probe_xp.get_text()
            # Use None if no previous value exists
            if xp == "":
                xp = None
            else:
                xp = float(xp)

        res = 0

        if xm is None or xp is None:
            return res

        if xm < xp:
            res = xp - xm
        else:
            res = xm - xp

        return res

    def length_y(self, ym=None, yp=None):
        """ Calculates a length in the Y direction """
        # Use previous value for ym if not supplied
        if ym is None:
            ym = self._lb_probe_ym.get_text()
            # Use None if no previous value exists
            if ym == "":
                ym = None
            else:
                ym = float(ym)

        # Use previous value for yp if not supplied
        if yp is None:
            yp = self._lb_probe_yp.get_text()
            # Use None if no previous value exists
            if yp == "":
                yp = None
            else:
                yp = float(yp)

        res = 0

        if ym is None or yp is None:
            return res

        if ym < yp:
            res = yp - ym
        else:
            res = ym - yp

        return res

    # --------------------------
    #
    #  Generic UI Methods
    #
    # --------------------------
    def on_common_spbtn_key_press_event(self, pin_name, gtkspinbutton, data=None):
        keyname = gtk.gdk.keyval_name(data.keyval)
        if keyname == "Return":
            # Drop the Italics
            gtkspinbutton.modify_font(pango.FontDescription("normal"))
        elif keyname == "Escape":
            # Restore the original value
            gtkspinbutton.set_value(self.halcomp[pin_name])

            # Drop the Italics
            gtkspinbutton.modify_font(pango.FontDescription("normal"))
        else:
            # Set to Italics
            gtkspinbutton.modify_font(pango.FontDescription("italic"))

    def on_common_spbtn_value_changed(
        self, pin_name, gtkspinbutton, data=None, _type=float
    ):
        # Drop the Italics
        gtkspinbutton.modify_font(pango.FontDescription("normal"))

        # Update the pin
        self.halcomp[pin_name] = gtkspinbutton.get_value()

        # Update the preferences
        self.prefs.putpref(pin_name, gtkspinbutton.get_value(), _type)

    # --------------------------
    #
    #  Generic Method Wrappers
    #
    # --------------------------
    @classmethod
    def ensure_errors_dismissed(cls, f):
        """ Ensures all errors have been dismissed, otherwise, shows a warning dialog """

        @wraps(f)
        def wrapper(self, *args, **kwargs):
            if self.halcomp["gui_error"]:
            # if self.error_poll() == -1:
                message = _("Please dismiss & act upon all errors")
                secondary = _("You can retry once done")
                self.warning_dialog(message, secondary=secondary)
                raise ValueError("error message on screen")
            
            self.stat.poll()
            prev_task_mode = self.stat.task_mode

            try:
                self.switch_to_mdi()
                self.probe_on()
                # Execute wrapped function
                return f(self, *args, **kwargs)
                
            except ValueError as msg:
                self.add_history_text(str(msg))
                #don't want to do this as we may want the probe on to fix what went wrong
                #self.probe_off()
                # Restore Previous Task Mode
                self.command.mode(prev_task_mode)
                self.wait_complete_non_blocking(5, "error restoring state")
                
            finally:
                # Restore Previous Task Mode
            #    self.probe_off()
                self.command.mode(prev_task_mode)
                self.wait_complete_non_blocking(5, "error restoring state")

        return wrapper
