#!/usr/bin/env python
#
# Copyright (c) 2015 Serguei Glavatski ( verser  from cnc-club.ru )
# Copyright (c) 2020 Probe Screen NG Developers
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import math

import hal
import hal_glib

import pylab as pl
import numpy as np

from .base import ProbeScreenBase


class ProbeScreenCalibration(ProbeScreenBase):
    # --------------------------
    #
    #  INIT
    #
    # --------------------------
    def __init__(self, halcomp, builder, useropts):
        super(ProbeScreenCalibration, self).__init__(halcomp, builder, useropts)

        self.spbtn_reni_channel = self.builder.get_object("spbtn_reni_channel")
        self.spbtn_reni_channel.set_value(1)

        self.btn_sphere_calibration = self.builder.get_object("sphere_calibration")
        self.btn_ring_calibration = self.builder.get_object("ring_calibration")

        self.spbt_sphere_diameter = self.builder.get_object("spbt_sphere_diameter")
        self.spbt_sphere_diameter.set_value(
            self.prefs.getpref("sphere_diameter", 1.0, float)
        )
        
        self.halcomp.newpin("sphere_diameter", hal.HAL_FLOAT, hal.HAL_OUT)
        self.halcomp["sphere_diameter"] = self.spbt_sphere_diameter.get_value()

        self.spbt_ring_diameter = self.builder.get_object("spbt_ring_diameter")
        self.spbt_ring_diameter.set_value(
            self.prefs.getpref("ring_diameter", 1.0, float)
        )
        self.halcomp.newpin("ring_diameter", hal.HAL_FLOAT, hal.HAL_OUT)
        self.halcomp["ring_diameter"] = self.spbt_ring_diameter.get_value()

        self.spbtn_num_points = self.builder.get_object("spbtn_num_points")
        self.spbtn_num_points.set_value(
            self.prefs.getpref("char_points", 2.0, float)
        )
        self.halcomp.newpin("char_points", hal.HAL_FLOAT, hal.HAL_OUT)
        self.halcomp["char_points"] = self.spbtn_num_points.get_value()

        self.spbtn_num_hits = self.builder.get_object("spbtn_num_hits")
        self.spbtn_num_hits.set_value(
            self.prefs.getpref("num_hits", 1.0, float)
        )
        self.halcomp.newpin("num_hits", hal.HAL_FLOAT, hal.HAL_OUT)
        self.halcomp["num_hits"] = self.spbtn_num_hits.get_value()
        
        self.tglbtn_ring_standard = self.builder.get_object("tglbtn_ring_standard")
        self.tglbtn_sphere_standard = self.builder.get_object("tglbtn_sphere_standard")
        self.tglbtn_ring_standard.set_active(True)
        
        self.chk_center_probe = self.builder.get_object("chk_center_probe")
        self.chk_center_probe.set_active(True)
        
        
    def on_spbt_sphere_diameter_key_press_event(self, gtkspinbutton, data=None):
        self.on_common_spbtn_key_press_event("sphere_diameter", gtkspinbutton, data)

    def on_spbt_sphere_diameter_value_changed(self, gtkspinbutton, data=None):
        self.on_common_spbtn_value_changed("sphere_diameter", gtkspinbutton, data)

    def on_spbt_ring_diameter_key_press_event(self, gtkspinbutton, data=None):
        self.on_common_spbtn_key_press_event("ring_diameter", gtkspinbutton, data)

    def on_spbt_ring_diameter_value_changed(self, gtkspinbutton, data=None):
        self.on_common_spbtn_value_changed("ring_diameter", gtkspinbutton, data)

    def on_spbtn_num_points_key_press_event(self, gtkspinbutton, data=None):
        self.on_common_spbtn_key_press_event("char_points", gtkspinbutton, data)

    def on_spbtn_num_points_value_changed(self, gtkspinbutton, data=None):
        self.on_common_spbtn_value_changed("char_points", gtkspinbutton, data)

    def on_spbtn_num_hits_key_press_event(self, gtkspinbutton, data=None):
        self.on_common_spbtn_key_press_event("num_hits", gtkspinbutton, data)

    def on_spbtn_num_hits_value_changed(self, gtkspinbutton, data=None):
        self.on_common_spbtn_value_changed("num_hits", gtkspinbutton, data)
        

    def on_tglbtn_ring_standard_toggled(self, gtktogglebutton, data=None):
        if self.tglbtn_ring_standard.get_active():
            self.tglbtn_sphere_standard.set_active(False)
    
    def on_tglbtn_sphere_standard_toggled(self, gtktogglebutton, data=None):
        if self.tglbtn_sphere_standard.get_active():
            self.tglbtn_ring_standard.set_active(False)
    

    @ProbeScreenBase.ensure_errors_dismissed
    def on_btn_start_probe_char_released(self, gtkbutton, data=None):
        if self.tglbtn_ring_standard.get_active():
            cal_diameter = self.spbt_ring_diameter.get_value()
            direction = -1
            zdown = False
        else:
            cal_diameter = self.spbt_sphere_diameter.get_value()
            direction = 1
            zdown = True
        
        numpoints = self.spbtn_num_points.get_value()
        numhits = self.spbtn_num_hits.get_value()
        
        b = self.current_position_with_offsets()
        
        if self.chk_center_probe.get_active():
            centervals = self.center_probe(direction, cal_diameter)
            z_offset = centervals[0]
            b[0] = centervals[1]
            b[1] = centervals[2]
        elif direction == 1:
            # find the top of the sphere and the z depth for the centerline
            ztouch = self.probe_z(-self.halcomp["ps_probe_max"])
            z_offset = (float(b[2]) - ztouch) + cal_diameter / 2 + self.halcomp["ps_probe_diam"] / 2 - self.halcomp["ps_z_clearance"]
            
        else:
            z_offset = 0
        
        results = self.probe_circle(
            dia_estimate = cal_diameter,
            num_points = numpoints, 
            direction = direction, 
            z_offset = z_offset,
            zdown = zdown, 
            orient_angle = 0, 
            orient_enable = True,
            num_samples = numhits,
            fine_probe = True,
            raw_output = True, 
            )
        
        # find the error values
        results[:] = [(x[0], x[1] - cal_diameter / 2 - self.halcomp["ps_probe_diam"] / 2) for x in results]
        
        self.gcode(self.move_xy(b[0], b[1]))
        
        self.add_history_text(str(results))
    
    @ProbeScreenBase.ensure_errors_dismissed
    def on_btn_reni_pair_released(self, gtkbutton, data=None):
        channel = self.spbtn_reni_channel.get_value()
        self.ocode("o<renikey> call [%d] [1]" % channel)
        self.add_history_text("RMI-Q started pairing on channel %d" % channel)
        
    @ProbeScreenBase.ensure_errors_dismissed
    def on_btn_reni_clear_released(self, gtkbutton, data=None):
        channel = self.spbtn_reni_channel.get_value()
        self.ocode("o<renikey> call [%d] [2]" % channel)
        self.add_history_text("RMI-Q cleared probe from channel %d" % channel)

    @ProbeScreenBase.ensure_errors_dismissed
    def on_btn_reni_clear_all_released(self, gtkbutton, data=None):
        self.ocode("o<renikey> call [1] [3]")
        self.add_history_text("RMI-Q cleared probes from all channels")

    # calibrate the probe using a ring gauge
    @ProbeScreenBase.ensure_errors_dismissed
    def on_ring_calibration_released(self, gtkbutton, data=None):
        cal_diameter = self.spbt_ring_diameter.get_value()
        self.probe_cal_ring(cal_diameter)
    
    # calibrate the probe using a sphere
    @ProbeScreenBase.ensure_errors_dismissed
    def on_sphere_calibration_released(self, gtkbutton, data=None):
        cal_diameter = self.spbt_sphere_diameter.get_value()
        self.probe_cal_sphere(cal_diameter)
    
    def center_probe(self, direction = 1, cal_diameter = 0.0):

        if cal_diameter <= 0.0:
            raise ValueError("calibration artifact diameter must be greater than zero")
        
        if direction == 1:
            b = self.current_position_with_offsets()
            # find the top of the sphere and the z depth for the centerline
            ztouch = self.probe_z(-self.halcomp["ps_probe_max"])
            zoff = (float(b[2]) - ztouch) + cal_diameter / 2 + self.halcomp["ps_probe_diam"] / 2 - self.halcomp["ps_z_clearance"]
            zdown = True
        else:
            zoff = 0
            zdown = False
        
        #rough center finding, just to make sure we don't have high tangent errors
        self.probe_xy(
            direction = direction,
            x_estimate = cal_diameter,
            y_estimate = cal_diameter,
            zdown = zdown,
            z_offset = zoff,
            move_center = True
        )
        
        if direction == 1:
            # accurately located the z depth now the probe center is aligned with the center of the standard
            b = self.current_position_with_offsets()
            ztouch = self.probe_z(-self.halcomp["ps_probe_max"])
            zoff = (float(b[2]) - ztouch) + cal_diameter / 2 + self.halcomp["ps_probe_diam"] / 2 - self.halcomp["ps_z_clearance"]
        
        # find the center of the sphere
        xp1, xn1, yp1, yn1 = self.probe_xy(
            direction = direction, 
            x_estimate = cal_diameter, 
            y_estimate = cal_diameter, 
            z_offset = zoff,
            zdown = zdown,
            orient_angle = 0, 
            orient_enable = True, 
            raw_output = True
        )
        # probe again but with the probe turned 180 degrees
        xp2, xn2, yp2, yn2 = self.probe_xy(
            direction = direction, 
            x_estimate = cal_diameter, 
            y_estimate = cal_diameter, 
            z_offset = zoff,
            zdown = zdown,
            orient_angle = 180, 
            orient_enable = True, 
            raw_output = True
        )
        
        #calculate center point, removed probe offsets
        xc = ((xp1 + xp2) / 2 + (xn1 + xn2) / 2) / 2
        yc = ((yp1 + yp2) / 2 + (yn1 + yn2) / 2) / 2
        
        #calculate the probe center offset
        xoff = ((xp1 - xp2) + (xn1 - xn2)) / 4
        yoff = ((yp1 - yp2) + (yn1 - yn2)) / 4
        
        self.add_history_text("Probe Centering results: x-center=%.5f, y-center=%.5f, x-offset=%.5f, y-offset=%.5f" % (xc, yc, xoff, yoff))
        
        # move to accurate center
        self.gcode(self.move_xy(xc, yc))
        
        return zoff, xc, yc, xoff, yoff

        
    
    def probe_cal_ring(self, cal_diameter = 0.0):
        
        s = ""
        zoff, xc, yc, xoff, yoff = self.center_probe(-1, cal_diameter)
        
        #now measure the diameter to find the probe diameter
        xp1, xn1, yp1, yn1 = self.probe_xy(
            direction = -1, 
            x_estimate = cal_diameter, 
            y_estimate = cal_diameter, 
            zdown = False, 
            orient_angle = 0, 
            orient_enable = True, 
            raw_output = True, 
            s = s
        )
        
        #average the x and y measurments
        probe_dia = (cal_diameter * 2 - (xp1 - xn1) - (yp1 - yn1)) / 2
        
        self.add_history_text("probe diameter = %.5f" % probe_dia)
        
        return xoff, yoff, probe_dia, xp1, xn1, yp1, yn1
       
    def probe_cal_sphere(self, cal_diameter = 0.0):
        
        s = ""
        #need to be accurately on the center of the sphere to have an accurate diameter
        zoff, xc, yc, xoff, yoff = self.center_probe(1, cal_diameter)
        
        # move to the probe center to get the diameter to remove probe offset effects
        self.gcode(self.move_rel_xy(xoff, yoff))
        
        xp1, xn1, yp1, yn1 = self.probe_xy(
            direction = 1, 
            x_estimate = cal_diameter, 
            y_estimate = cal_diameter, 
            z_offset = zoff, 
            orient_angle = 0, 
            orient_enable = True, 
            raw_output = True 
        )
        
        #average the x and y measurments
        probe_dia = (((xp1 - xn1) + (yp1 - yn1)) - cal_diameter * 2 ) / 2

        self.gcode(self.move_xy(xc, yc))
        
        self.add_history_text("probe diameter = %f" % probe_dia)
        
        return xoff, yoff, probe_dia, xp1, xn1, yp1, yn1

    def probe_sphere_circle(self, 
        cal_diameter,
        sphere_center,
        num_points, 
        direction = 1, 
        z_offset = 0,
        z_angle = 0,
        zdown = True, 
        orient_angle = 0, 
        orient_enable = True,
        fine_probe = False,
        raw_output = False, 
        s = ""
    ):
        if num_points < 2:
            raise ValueError("Must have at least 2 points to probe a circle")
            
        #capture the start point for use
        startpoint = self.current_position_with_offsets()
        
        if orient_enable:
            s += self.spindle_orient(orient_angle)
        
        # calculate clearances
        if fine_probe:
            backoff = self.halcomp["ps_probe_latch"] + self.halcomp["ps_probe_diam"] / 2
        else:
            backoff = self.halcomp["ps_xy_clearance"]
            
        if direction == 1:
            diaclear = cal_diameter / 2 + backoff
        else:
            if (cal_diameter / 2) < backoff:
                diaclear = 0
            else:
                diaclear = cal_diameter / 2 - backoff
                
        s += self.move_rel_xy(diaclear, 0)
        
        if direction == 1 or zdown:
            s += self.z_clearance_down(z_offset)
        
        self.gcode(s)
        s = ""
        
        radius = diaclear
        
        lengths = []
        
        for x in range(int(num_points)):
            a = self.current_position_with_offsets()
            current_angle = 360.0 / float(num_points) * float(x)
            probe_angle = current_angle
            if direction == 1:
                probe_angle += 180.0
                if probe_angle > 360.0:
                    probe_angle -= 360.0
            
            if fine_probe:
                xt, yt, zt = self.probe_fine(probe_angle, z_angle)
            else:
                xt, yt, zt = self.probe_raw(probe_angle, z_angle)
                
            length = ((xt - float(sphere_center[0])) ** 2 + (yt - float(sphere_center[1])) ** 2 + (zt - sphere_center[2]) ** 2) ** .5
            if not raw_output:
                length += 0.5 * -1 * direction * self.halcomp["ps_probe_diam"]
            lengths.append((probe_angle / 180.0 * math.pi, length))
            self.add_history_text("angle: %.3f  length: %.5f" % (probe_angle, length))
            # return to probe start point
            s = self.move_xyz(float(a[0]), float(a[1]), float(a[2]))
            
            # move in an arc to the next point
            if radius > 0:
                next_angle = 360.0 / float(num_points) * float(x + 1)
                next_x = math.cos(next_angle/180.0 * math.pi) * radius + float(sphere_center[0])
                next_y = math.sin(next_angle/180.0 * math.pi) * radius + float(sphere_center[1])
                i_val = float(sphere_center[0]) - float(a[0])
                j_val = float(sphere_center[1]) - float(a[1])
                s += "G3 X%f Y%f I%f J%f\n" % (next_x, next_y, i_val, j_val)
                
            self.gcode(s)
            s = ""
            
        if direction == 1 or zdown:
            s = self.z_clearance_up(z_offset)
        
        self.gcode(s)
        
        return lengths

                
